@@ -190,7 +190,7 @@ def _build_index_to_cityles_map(land_cover_source):
 
 
 def export_topog(building_height_grid, building_id_grid, output_path, 
-                 building_material='default'):
+                 building_material='default', cityles_landuse_grid=None):
     """
     Export topog.txt file for CityLES
     
@@ -211,8 +211,9 @@ def export_topog(building_height_grid, building_id_grid, output_path,
     material_code = BUILDING_MATERIAL_MAPPING.get(building_material, 
                                                   BUILDING_MATERIAL_MAPPING['default'])
     
-    # Write all grid cells including those without buildings
-    n_buildings = ny * nx
+    # Count only cells with building height > 0
+    building_mask = building_height_grid > 0
+    n_buildings = int(np.count_nonzero(building_mask))
     
     with open(filename, 'w') as f:
         # Write number of buildings
@@ -225,8 +226,17 @@ def export_topog(building_height_grid, building_id_grid, output_path,
                 i_1based = i + 1
                 j_1based = j + 1
                 height = float(building_height_grid[j, i])
+                # Decide material code per cell
+                if cityles_landuse_grid is not None:
+                    cell_lu = int(cityles_landuse_grid[j, i])
+                    material_code_cell = cell_lu + 100
+                else:
+                    if height > 0:
+                        material_code_cell = material_code
+                    else:
+                        material_code_cell = 102
                 # Format: i j height material_code depth1 depth2 changed_material
-                f.write(f"{i_1based} {j_1based} {height:.1f} {material_code} 0.0 0.0 102\n")
+                f.write(f"{i_1based} {j_1based} {height:.1f} {material_code_cell} 0.0 0.0 102\n")
 
 
 def export_landuse(land_cover_grid, output_path, land_cover_source=None):
@@ -253,6 +263,8 @@ def export_landuse(land_cover_grid, output_path, land_cover_source=None):
 
     # Create mapping statistics
     mapping_stats = {}
+    # Prepare grid to return
+    cityles_landuse_grid = np.zeros((ny, nx), dtype=int)
 
     with open(filename, 'w') as f:
         # Write in row-major order (j varies first, then i)
@@ -262,6 +274,8 @@ def export_landuse(land_cover_grid, output_path, land_cover_source=None):
                 cityles_code = index_to_code.get(idx, 4)
                 f.write(f"{cityles_code}\n")
 
+                cityles_landuse_grid[j, i] = cityles_code
+
                 # Track mapping statistics
                 if idx not in mapping_stats:
                     mapping_stats[idx] = {'cityles_code': cityles_code, 'count': 0}
@@ -276,6 +290,8 @@ def export_landuse(land_cover_grid, output_path, land_cover_source=None):
         class_name = class_names[idx] if 0 <= idx < len(class_names) else 'Unknown'
         print(f"  {idx}: {class_name} -> CityLES {stats['cityles_code']}: "
               f"{stats['count']} cells ({percentage:.1f}%)")
+    
+    return cityles_landuse_grid
 
 
 def export_dem(dem_grid, output_path):
@@ -299,12 +315,14 @@ def export_dem(dem_grid, output_path):
                 # CityLES uses 1-based indexing
                 i_1based = i + 1
                 j_1based = j + 1
-                elevation = dem_grid[j, i]
-                
+                elevation = float(dem_grid[j, i])
+                # Clamp negative elevations to 0.0 meters
+                if elevation < 0.0:
+                    elevation = 0.0
                 f.write(f"{i_1based} {j_1based} {elevation:.1f}\n")
 
 
-def export_vmap(canopy_height_grid, output_path, tree_base_ratio=0.3, tree_type='default'):
+def export_vmap(canopy_height_grid, output_path, tree_base_ratio=0.3, tree_type='default', building_height_grid=None, canopy_bottom_height_grid=None):
     """
     Export vmap.txt file for CityLES
     
@@ -324,8 +342,15 @@ def export_vmap(canopy_height_grid, output_path, tree_base_ratio=0.3, tree_type=
     ny, nx = canopy_height_grid.shape
     tree_code = TREE_TYPE_MAPPING.get(tree_type, TREE_TYPE_MAPPING['default'])
     
-    # Write all grid cells including those without vegetation
-    n_trees = ny * nx
+    # If building heights are provided, remove trees where buildings exist
+    if building_height_grid is not None:
+        effective_canopy = np.where(building_height_grid > 0, 0.0, canopy_height_grid)
+    else:
+        effective_canopy = canopy_height_grid
+    
+    # Count only cells with canopy height > 0
+    vegetation_mask = effective_canopy > 0
+    n_trees = int(np.count_nonzero(vegetation_mask))
     
     with open(filename, 'w') as f:
         # Write number of trees
@@ -337,8 +362,11 @@ def export_vmap(canopy_height_grid, output_path, tree_base_ratio=0.3, tree_type=
                 # CityLES uses 1-based indexing
                 i_1based = i + 1
                 j_1based = j + 1
-                total_height = float(canopy_height_grid[j, i])
-                lower_height = total_height * tree_base_ratio
+                total_height = float(effective_canopy[j, i])
+                if canopy_bottom_height_grid is not None:
+                    lower_height = float(np.clip(canopy_bottom_height_grid[j, i], 0.0, total_height))
+                else:
+                    lower_height = total_height * tree_base_ratio
                 upper_height = total_height
                 # Format: i j lower_height upper_height tree_type
                 f.write(f"{i_1based} {j_1based} {lower_height:.1f} {upper_height:.1f} {tree_code}\n")
@@ -388,7 +416,7 @@ def export_cityles(building_height_grid, building_id_grid, canopy_height_grid,
                    land_cover_grid, dem_grid, meshsize, land_cover_source,
                    rectangle_vertices, output_directory="output/cityles",
                    building_material='default', tree_type='default',
-                   tree_base_ratio=0.3, **kwargs):
+                   tree_base_ratio=0.3, canopy_bottom_height_grid=None, **kwargs):
     """
     Export VoxCity data to CityLES format
     
@@ -432,17 +460,23 @@ def export_cityles(building_height_grid, building_id_grid, canopy_height_grid,
     print(f"Land cover source: {land_cover_source}")
     
     # Export individual files
-    print("\nExporting topog.txt...")
-    export_topog(building_height_grid, building_id_grid, output_path, building_material)
-    
     print("\nExporting landuse.txt...")
-    export_landuse(land_cover_grid, output_path, land_cover_source)
+    cityles_landuse_grid = export_landuse(land_cover_grid, output_path, land_cover_source)
+
+    print("\nExporting topog.txt...")
+    export_topog(
+        building_height_grid,
+        building_id_grid,
+        output_path,
+        building_material,
+        cityles_landuse_grid=cityles_landuse_grid,
+    )
     
     print("\nExporting dem.txt...")
     export_dem(dem_grid, output_path)
     
     print("\nExporting vmap.txt...")
-    export_vmap(canopy_height_grid, output_path, tree_base_ratio, tree_type)
+    export_vmap(canopy_height_grid, output_path, tree_base_ratio, tree_type, building_height_grid=building_height_grid, canopy_bottom_height_grid=canopy_bottom_height_grid)
     
     print("\nExporting lonlat.txt...")
     export_lonlat(rectangle_vertices, building_height_grid.shape, output_path)
@@ -460,7 +494,9 @@ def export_cityles(building_height_grid, building_id_grid, canopy_height_grid,
         f.write(f"Tree type: {tree_type}\n")
         f.write(f"Bounds: {rectangle_vertices}\n")
         f.write(f"Buildings: {np.sum(building_height_grid > 0)}\n")
-        f.write(f"Trees: {np.sum(canopy_height_grid > 0)}\n")
+        # Trees count after removing overlaps with buildings
+        trees_count = int(np.sum(np.where(building_height_grid > 0, 0.0, canopy_height_grid) > 0))
+        f.write(f"Trees: {trees_count}\n")
         
         # Add land use value ranges
         f.write(f"\nLand cover value range: {land_cover_grid.min()} - {land_cover_grid.max()}\n")
