@@ -75,6 +75,7 @@ from .geoprocessor.grid import (
     create_land_cover_grid_from_gdf_polygon,
     create_building_height_grid_from_open_building_temporal_polygon,
     create_vegetation_height_grid_from_gdf_polygon,
+    create_canopy_grids_from_tree_gdf,
     create_dem_grid_from_gdf_polygon
 )
 
@@ -225,12 +226,14 @@ def get_building_height_grid(rectangle_vertices, meshsize, source, output_dir, b
     else:
         # Fetch building data from primary source
         # Each source has different data formats and processing requirements
+        # Floor height (m) for inferring heights from floors/levels
+        floor_height = kwargs.get("floor_height", 3.0)
         if source == 'Microsoft Building Footprints':
             # Machine learning-derived building footprints from satellite imagery
             gdf = get_mbfp_gdf(output_dir, rectangle_vertices)
         elif source == 'OpenStreetMap':
             # Crowd-sourced building data with varying completeness
-            gdf = load_gdf_from_openstreetmap(rectangle_vertices)
+            gdf = load_gdf_from_openstreetmap(rectangle_vertices, floor_height=floor_height)
         elif source == "Open Building 2.5D Temporal":
             # Special case: this source provides both footprints and heights
             # Skip GeoDataFrame processing and create grids directly
@@ -243,7 +246,7 @@ def get_building_height_grid(rectangle_vertices, meshsize, source, output_dir, b
         #     gdf = load_gdf_from_openmaptiles(rectangle_vertices, kwargs["maptiler_API_key"])
         elif source == "Overture":
             # Open building dataset from Overture Maps Foundation
-            gdf = load_gdf_from_overture(rectangle_vertices)
+            gdf = load_gdf_from_overture(rectangle_vertices, floor_height=floor_height)
         elif source == "Local file":
             # Handle user-provided local building data files
             _, extension = os.path.splitext(kwargs["building_path"])
@@ -284,13 +287,13 @@ def get_building_height_grid(rectangle_vertices, meshsize, source, output_dir, b
             if building_complementary_source == 'Microsoft Building Footprints':
                 gdf_comp = get_mbfp_gdf(output_dir, rectangle_vertices)
             elif building_complementary_source == 'OpenStreetMap':
-                gdf_comp = load_gdf_from_openstreetmap(rectangle_vertices)
+                gdf_comp = load_gdf_from_openstreetmap(rectangle_vertices, floor_height=floor_height)
             elif building_complementary_source == 'EUBUCCO v0.1':
                 gdf_comp = load_gdf_from_eubucco(rectangle_vertices, output_dir)
             # elif building_complementary_source == "OpenMapTiles":
             #     gdf_comp = load_gdf_from_openmaptiles(rectangle_vertices, kwargs["maptiler_API_key"])
             elif building_complementary_source == "Overture":
-                gdf_comp = load_gdf_from_overture(rectangle_vertices)
+                gdf_comp = load_gdf_from_overture(rectangle_vertices, floor_height=floor_height)
             elif building_complementary_source == "Local file":
                 _, extension = os.path.splitext(kwargs["building_complementary_path"])
                 if extension == ".gpkg":
@@ -313,58 +316,85 @@ def get_building_height_grid(rectangle_vertices, meshsize, source, output_dir, b
     return building_height_grid, building_min_height_grid, building_id_grid, filtered_buildings
 
 def get_canopy_height_grid(rectangle_vertices, meshsize, source, output_dir, **kwargs):
-    """Creates a grid of tree canopy heights.
+    """Creates canopy top and bottom height grids.
+
+    Supports satellite sources and a GeoDataFrame source (from draw_additional_trees).
 
     Args:
         rectangle_vertices: List of coordinates defining the area of interest
         meshsize: Size of each grid cell in meters
-        source: Data source for canopy heights
+        source: Data source for canopy heights. Use 'GeoDataFrame' or 'tree_gdf' for tree_gdf path/object.
         output_dir: Directory to save output files
         **kwargs: Additional arguments including:
             - gridvis: Whether to visualize the grid
+            - tree_gdf: GeoDataFrame of trees (optional when source='GeoDataFrame')
+            - tree_gdf_path: Path to a local file (e.g., .gpkg) with trees when source='GeoDataFrame'
+            - trunk_height_ratio: Fraction of top height used as canopy bottom for non-GDF sources
 
     Returns:
-        numpy.ndarray: Grid of canopy heights
+        tuple[numpy.ndarray, numpy.ndarray]: (canopy_top_height_grid, canopy_bottom_height_grid)
     """
 
     print("Creating Canopy Height grid\n ")
-    print(f"Data source: High Resolution Canopy Height Maps by WRI and Meta")
-    
-    # Initialize Earth Engine for satellite-based canopy height data
-    initialize_earth_engine()
+    print(f"Data source: {source}")
 
     os.makedirs(output_dir, exist_ok=True)
+
+    # Branch: compute from a provided GeoDataFrame (draw_additional_trees output)
+    if source in ('GeoDataFrame', 'tree_gdf', 'Tree_GeoDataFrame', 'GDF'):
+        tree_gdf = kwargs.get('tree_gdf')
+        tree_gdf_path = kwargs.get('tree_gdf_path')
+        if tree_gdf is None and tree_gdf_path is not None:
+            _, ext = os.path.splitext(tree_gdf_path)
+            if ext.lower() == '.gpkg':
+                tree_gdf = get_gdf_from_gpkg(tree_gdf_path, rectangle_vertices)
+            else:
+                raise ValueError("Unsupported tree file format. Use .gpkg or pass a GeoDataFrame.")
+        if tree_gdf is None:
+            raise ValueError("When source='GeoDataFrame', provide 'tree_gdf' or 'tree_gdf_path'.")
+
+        canopy_top, canopy_bottom = create_canopy_grids_from_tree_gdf(tree_gdf, meshsize, rectangle_vertices)
+
+        # Visualization
+        grid_vis = kwargs.get("gridvis", True)
+        if grid_vis:
+            vis = canopy_top.copy()
+            vis[vis == 0] = np.nan
+            visualize_numerical_grid(np.flipud(vis), meshsize, "Tree canopy height (top)", cmap='Greens', label='Tree canopy height (m)')
+
+        return canopy_top, canopy_bottom
+
+    # Default: satellite/remote sensing sources
+    print("Data source: High Resolution Canopy Height Maps by WRI and Meta")
+    initialize_earth_engine()
+
     geotiff_path = os.path.join(output_dir, "canopy_height.tif")
-    
-    # Get region of interest and fetch canopy height data
+
     roi = get_roi(rectangle_vertices)
     if source == 'High Resolution 1m Global Canopy Height Maps':
-        # High-resolution (1m) global canopy height maps from Meta and WRI
-        # Based on satellite imagery and machine learning models
-        collection_name = "projects/meta-forest-monitoring-okw37/assets/CanopyHeight"  
-        image = get_ee_image_collection(collection_name, roi)      
+        collection_name = "projects/meta-forest-monitoring-okw37/assets/CanopyHeight"
+        image = get_ee_image_collection(collection_name, roi)
     elif source == 'ETH Global Sentinel-2 10m Canopy Height (2020)':
-        # Medium-resolution (10m) canopy height from ETH Zurich
-        # Derived from Sentinel-2 satellite data
         collection_name = "users/nlang/ETH_GlobalCanopyHeight_2020_10m_v1"
         image = get_ee_image(collection_name, roi)
-    
-    # Save canopy height data as GeoTIFF with specified resolution
-    save_geotiff(image, geotiff_path, resolution=meshsize)  
+    else:
+        raise ValueError(f"Unsupported canopy source: {source}")
 
-    # Convert GeoTIFF to regular grid format
+    save_geotiff(image, geotiff_path, resolution=meshsize)
     canopy_height_grid = create_height_grid_from_geotiff_polygon(geotiff_path, meshsize, rectangle_vertices)
 
-    # Generate visualization if requested
-    grid_vis = kwargs.get("gridvis", True)    
+    # Derive bottom grid using trunk_height_ratio (consistent with create_3d_voxel)
+    trunk_height_ratio = kwargs.get("trunk_height_ratio")
+    if trunk_height_ratio is None:
+        trunk_height_ratio = 11.76 / 19.98
+    canopy_bottom_grid = canopy_height_grid * float(trunk_height_ratio)
+
+    grid_vis = kwargs.get("gridvis", True)
     if grid_vis:
-        # Replace zeros with NaN for better visualization (show only areas with trees)
         canopy_height_grid_nan = canopy_height_grid.copy()
         canopy_height_grid_nan[canopy_height_grid_nan == 0] = np.nan
-        # Use green color scheme appropriate for vegetation
         visualize_numerical_grid(np.flipud(canopy_height_grid_nan), meshsize, "Tree canopy height", cmap='Greens', label='Tree canopy height (m)')
-
-    return canopy_height_grid
+    return canopy_height_grid, canopy_bottom_grid
 
 def get_dem_grid(rectangle_vertices, meshsize, source, output_dir, **kwargs):
     """Creates a digital elevation model grid.
@@ -431,7 +461,7 @@ def get_dem_grid(rectangle_vertices, meshsize, source, output_dir, **kwargs):
 
 def create_3d_voxel(building_height_grid_ori, building_min_height_grid_ori, 
                     building_id_grid_ori, land_cover_grid_ori, dem_grid_ori, 
-                    tree_grid_ori, voxel_size, land_cover_source, **kwargs):
+                    tree_grid_ori, voxel_size, land_cover_source, canopy_bottom_height_grid_ori=None, **kwargs):
     """Creates a 3D voxel representation combining all input grids.
     Args:
         building_height_grid_ori: Grid of building heights
@@ -465,6 +495,9 @@ def create_3d_voxel(building_height_grid_ori, building_min_height_grid_ori,
     dem_grid = np.flipud(dem_grid_ori.copy()) - np.min(dem_grid_ori) # Normalize DEM to start at 0
     dem_grid = process_grid(building_id_grid, dem_grid) # Process DEM based on building footprints
     tree_grid = np.flipud(tree_grid_ori.copy())
+    canopy_bottom_grid = None
+    if canopy_bottom_height_grid_ori is not None:
+        canopy_bottom_grid = np.flipud(canopy_bottom_height_grid_ori.copy())
     # Validate that all input grids have consistent dimensions
     assert building_height_grid.shape == land_cover_grid.shape == dem_grid.shape == tree_grid.shape, "Input grids must have the same shape"
     rows, cols = building_height_grid.shape
@@ -509,7 +542,10 @@ def create_3d_voxel(building_height_grid_ori, building_min_height_grid_ori,
             # Process tree canopy if trees are present
             if tree_height > 0:
                 # Calculate tree structure: trunk base to crown base to crown top
-                crown_base_height = (tree_height * trunk_height_ratio)
+                if canopy_bottom_grid is not None:
+                    crown_base_height = canopy_bottom_grid[i, j]
+                else:
+                    crown_base_height = (tree_height * trunk_height_ratio)
                 crown_base_height_level = int(crown_base_height / voxel_size + 0.5)
                 crown_top_height = tree_height
                 crown_top_height_level = int(crown_top_height / voxel_size + 0.5)
@@ -648,7 +684,8 @@ def get_voxcity(rectangle_vertices, building_source, land_cover_source, canopy_h
             - building_height_grid: 2D grid of building heights
             - building_min_height_grid: 2D grid of minimum building heights
             - building_id_grid: 2D grid of building IDs
-            - canopy_height_grid: 2D grid of tree canopy heights
+            - canopy_height_grid: 2D grid of tree canopy top heights
+            - canopy_bottom_height_grid: 2D grid of tree canopy bottom heights
             - land_cover_grid: 2D grid of land cover classifications
             - dem_grid: 2D grid of ground elevation
             - building_geojson: GeoJSON of building footprints and metadata
@@ -677,19 +714,25 @@ def get_voxcity(rectangle_vertices, building_source, land_cover_source, canopy_h
     # STEP 2: Handle canopy height data
     # Either use static values or fetch from satellite sources
     if canopy_height_source == "Static":
-        # Create uniform canopy height for all tree-covered areas
+        # Create uniform canopy height for all tree-covered areas (top grid)
         canopy_height_grid = np.zeros_like(land_cover_grid, dtype=float)
-        
-        # Apply static height to areas classified as trees
-        # Default height represents typical urban tree height
         static_tree_height = kwargs.get("static_tree_height", 10.0)
-        tree_mask = (land_cover_grid == 4)
-        
-        # Set static height for tree cells
+        # Determine tree class indices based on source-specific class names
+        _classes = get_land_cover_classes(land_cover_source)
+        _class_to_int = {name: i for i, name in enumerate(_classes.values())}
+        _tree_labels = ["Tree", "Trees", "Tree Canopy"]
+        _tree_indices = [_class_to_int[label] for label in _tree_labels if label in _class_to_int]
+        tree_mask = np.isin(land_cover_grid, _tree_indices) if _tree_indices else np.zeros_like(land_cover_grid, dtype=bool)
         canopy_height_grid[tree_mask] = static_tree_height
+
+        # Derive bottom from trunk_height_ratio
+        trunk_height_ratio = kwargs.get("trunk_height_ratio")
+        if trunk_height_ratio is None:
+            trunk_height_ratio = 11.76 / 19.98
+        canopy_bottom_height_grid = canopy_height_grid * float(trunk_height_ratio)
     else:
-        # Fetch canopy height from satellite/remote sensing sources
-        canopy_height_grid = get_canopy_height_grid(rectangle_vertices, meshsize, canopy_height_source, output_dir, **kwargs)
+        # Fetch canopy top/bottom from source
+        canopy_height_grid, canopy_bottom_height_grid = get_canopy_height_grid(rectangle_vertices, meshsize, canopy_height_source, output_dir, **kwargs)
     
     # STEP 3: Handle digital elevation model (terrain)
     if dem_source == "Flat":
@@ -705,17 +748,22 @@ def get_voxcity(rectangle_vertices, building_source, land_cover_source, canopy_h
     min_canopy_height = kwargs.get("min_canopy_height")
     if min_canopy_height is not None:
         canopy_height_grid[canopy_height_grid < kwargs["min_canopy_height"]] = 0        
+        if 'canopy_bottom_height_grid' in locals():
+            canopy_bottom_height_grid[canopy_height_grid == 0] = 0
 
     # Remove objects near the boundary to avoid edge effects
     # This is useful when the area of interest is part of a larger urban area
     remove_perimeter_object = kwargs.get("remove_perimeter_object")
     if (remove_perimeter_object is not None) and (remove_perimeter_object > 0):
+        print("apply perimeter removal")
         # Calculate perimeter width based on grid dimensions
         w_peri = int(remove_perimeter_object * building_height_grid.shape[0] + 0.5)
         h_peri = int(remove_perimeter_object * building_height_grid.shape[1] + 0.5)
         
         # Clear canopy heights in perimeter areas
         canopy_height_grid[:w_peri, :] = canopy_height_grid[-w_peri:, :] = canopy_height_grid[:, :h_peri] = canopy_height_grid[:, -h_peri:] = 0
+        if 'canopy_bottom_height_grid' in locals():
+            canopy_bottom_height_grid[:w_peri, :] = canopy_bottom_height_grid[-w_peri:, :] = canopy_bottom_height_grid[:, :h_peri] = canopy_bottom_height_grid[:, -h_peri:] = 0
 
         # Identify buildings that intersect with perimeter areas
         ids1 = np.unique(building_id_grid[:w_peri, :][building_id_grid[:w_peri, :] > 0])
@@ -730,6 +778,31 @@ def get_voxcity(rectangle_vertices, building_source, land_cover_source, canopy_h
             building_height_grid[positions] = 0
             building_min_height_grid[positions] = [[] for _ in range(len(building_min_height_grid[positions]))]
 
+        # Visualize grids after optional perimeter removal
+        grid_vis = kwargs.get("gridvis", True)
+        if grid_vis:
+            # Building height grid visualization (zeros hidden)
+            building_height_grid_nan = building_height_grid.copy()
+            building_height_grid_nan[building_height_grid_nan == 0] = np.nan
+            visualize_numerical_grid(
+                np.flipud(building_height_grid_nan),
+                meshsize,
+                "building height (m)",
+                cmap='viridis',
+                label='Value'
+            )
+
+            # Canopy height grid visualization (zeros hidden)
+            canopy_height_grid_nan = canopy_height_grid.copy()
+            canopy_height_grid_nan[canopy_height_grid_nan == 0] = np.nan
+            visualize_numerical_grid(
+                np.flipud(canopy_height_grid_nan),
+                meshsize,
+                "Tree canopy height (m)",
+                cmap='Greens',
+                label='Tree canopy height (m)'
+            )
+
     # STEP 5: Generate optional 2D visualizations on interactive maps
     mapvis = kwargs.get("mapvis")
     if mapvis:
@@ -789,7 +862,7 @@ def get_voxcity(rectangle_vertices, building_source, land_cover_source, canopy_h
 
     # STEP 6: Generate the final 3D voxel model
     # This combines all 2D grids into a comprehensive 3D representation
-    voxcity_grid = create_3d_voxel(building_height_grid, building_min_height_grid, building_id_grid, land_cover_grid, dem_grid, canopy_height_grid, meshsize, land_cover_source)
+    voxcity_grid = create_3d_voxel(building_height_grid, building_min_height_grid, building_id_grid, land_cover_grid, dem_grid, canopy_height_grid, meshsize, land_cover_source, canopy_bottom_height_grid)
 
     # STEP 7: Generate optional 3D visualization
     voxelvis = kwargs.get("voxelvis")
@@ -810,7 +883,7 @@ def get_voxcity(rectangle_vertices, building_source, land_cover_source, canopy_h
                          building_id_grid, canopy_height_grid, land_cover_grid, dem_grid, 
                          building_gdf, meshsize, rectangle_vertices)
     
-    return voxcity_grid, building_height_grid, building_min_height_grid, building_id_grid, canopy_height_grid, land_cover_grid, dem_grid, building_gdf
+    return voxcity_grid, building_height_grid, building_min_height_grid, building_id_grid, canopy_height_grid, canopy_bottom_height_grid, land_cover_grid, dem_grid, building_gdf
 
 def get_voxcity_CityGML(rectangle_vertices, land_cover_source, canopy_height_source, meshsize, url_citygml=None, citygml_path=None, **kwargs):
     """Main function to generate a complete voxel city model.
@@ -851,11 +924,105 @@ def get_voxcity_CityGML(rectangle_vertices, land_cover_source, canopy_height_sou
     # get all required gdfs    
     building_gdf, terrain_gdf, vegetation_gdf = load_buid_dem_veg_from_citygml(url=url_citygml, citygml_path=citygml_path, base_dir=output_dir, rectangle_vertices=rectangle_vertices)
 
+    # Normalize CRS to WGS84 (EPSG:4326) to ensure consistent operations downstream
+    try:
+        import geopandas as gpd  # noqa: F401
+        if building_gdf is not None:
+            if building_gdf.crs is None:
+                building_gdf = building_gdf.set_crs(epsg=4326)
+            elif getattr(building_gdf.crs, 'to_epsg', lambda: None)() != 4326 and building_gdf.crs != "EPSG:4326":
+                building_gdf = building_gdf.to_crs(epsg=4326)
+        if terrain_gdf is not None:
+            if terrain_gdf.crs is None:
+                terrain_gdf = terrain_gdf.set_crs(epsg=4326)
+            elif getattr(terrain_gdf.crs, 'to_epsg', lambda: None)() != 4326 and terrain_gdf.crs != "EPSG:4326":
+                terrain_gdf = terrain_gdf.to_crs(epsg=4326)
+        if vegetation_gdf is not None:
+            if vegetation_gdf.crs is None:
+                vegetation_gdf = vegetation_gdf.set_crs(epsg=4326)
+            elif getattr(vegetation_gdf.crs, 'to_epsg', lambda: None)() != 4326 and vegetation_gdf.crs != "EPSG:4326":
+                vegetation_gdf = vegetation_gdf.to_crs(epsg=4326)
+    except Exception:
+        pass
+
     land_cover_grid = get_land_cover_grid(rectangle_vertices, meshsize, land_cover_source, output_dir, **kwargs)        
 
     # building_height_grid, building_min_height_grid, building_id_grid, building_gdf = get_building_height_grid(rectangle_vertices, meshsize, building_source, output_dir, **kwargs)
     print("Creating building height grid")
-    building_height_grid, building_min_height_grid, building_id_grid, filtered_buildings = create_building_height_grid_from_gdf_polygon(building_gdf, meshsize, rectangle_vertices, **kwargs)
+    # Prepare complementary building source if provided
+    building_complementary_source = kwargs.get("building_complementary_source")
+    gdf_comp = None
+    geotiff_path_comp = None
+    complement_building_footprints = kwargs.get("complement_building_footprints")
+    # Default to complement footprints when a complementary source is specified
+    if complement_building_footprints is None and (building_complementary_source not in (None, "None")):
+        complement_building_footprints = True
+
+    if (building_complementary_source is not None) and (building_complementary_source != "None"):
+        # Vector complementary sources
+        floor_height = kwargs.get("floor_height", 3.0)
+        if building_complementary_source == 'Microsoft Building Footprints':
+            gdf_comp = get_mbfp_gdf(kwargs.get("output_dir", "output"), rectangle_vertices)
+        elif building_complementary_source == 'OpenStreetMap':
+            gdf_comp = load_gdf_from_openstreetmap(rectangle_vertices, floor_height=floor_height)
+        elif building_complementary_source == 'EUBUCCO v0.1':
+            gdf_comp = load_gdf_from_eubucco(rectangle_vertices, kwargs.get("output_dir", "output"))
+        elif building_complementary_source == 'Overture':
+            gdf_comp = load_gdf_from_overture(rectangle_vertices, floor_height=floor_height)
+        elif building_complementary_source == 'Local file':
+            comp_path = kwargs.get("building_complementary_path")
+            if comp_path is not None:
+                _, extension = os.path.splitext(comp_path)
+                if extension == ".gpkg":
+                    gdf_comp = get_gdf_from_gpkg(comp_path, rectangle_vertices)
+        # Ensure complementary GDF uses WGS84
+        if gdf_comp is not None:
+            try:
+                if gdf_comp.crs is None:
+                    gdf_comp = gdf_comp.set_crs(epsg=4326)
+                elif getattr(gdf_comp.crs, 'to_epsg', lambda: None)() != 4326 and gdf_comp.crs != "EPSG:4326":
+                    gdf_comp = gdf_comp.to_crs(epsg=4326)
+            except Exception:
+                pass
+        # Raster complementary sources (height only)
+        elif building_complementary_source == "Open Building 2.5D Temporal":
+            roi = get_roi(rectangle_vertices)
+            os.makedirs(kwargs.get("output_dir", "output"), exist_ok=True)
+            geotiff_path_comp = os.path.join(kwargs.get("output_dir", "output"), "building_height.tif")
+            save_geotiff_open_buildings_temporal(roi, geotiff_path_comp)
+        elif building_complementary_source in ["England 1m DSM - DTM", "Netherlands 0.5m DSM - DTM"]:
+            roi = get_roi(rectangle_vertices)
+            os.makedirs(kwargs.get("output_dir", "output"), exist_ok=True)
+            geotiff_path_comp = os.path.join(kwargs.get("output_dir", "output"), "building_height.tif")
+            save_geotiff_dsm_minus_dtm(roi, geotiff_path_comp, meshsize, building_complementary_source)
+
+    # Filter and assemble kwargs accepted by the grid function
+    _allowed_building_kwargs = {
+        "overlapping_footprint",
+        "gdf_comp",
+        "geotiff_path_comp",
+        "complement_building_footprints",
+        "complement_height",
+    }
+    _building_kwargs = {k: v for k, v in kwargs.items() if k in _allowed_building_kwargs}
+    if gdf_comp is not None:
+        _building_kwargs["gdf_comp"] = gdf_comp
+    if geotiff_path_comp is not None:
+        _building_kwargs["geotiff_path_comp"] = geotiff_path_comp
+    if complement_building_footprints is not None:
+        _building_kwargs["complement_building_footprints"] = complement_building_footprints
+
+    # Map user-provided building_complement_height -> complement_height for grid builder
+    comp_height_user = kwargs.get("building_complement_height")
+    if comp_height_user is not None:
+        _building_kwargs["complement_height"] = comp_height_user
+    # If footprints are being complemented and no height provided, default to 10
+    if _building_kwargs.get("complement_building_footprints") and ("complement_height" not in _building_kwargs):
+        _building_kwargs["complement_height"] = 10.0
+
+    building_height_grid, building_min_height_grid, building_id_grid, filtered_buildings = create_building_height_grid_from_gdf_polygon(
+        building_gdf, meshsize, rectangle_vertices, **_building_kwargs
+    )
 
     # Visualize grid if requested
     grid_vis = kwargs.get("gridvis", True)    
@@ -876,23 +1043,46 @@ def get_voxcity_CityGML(rectangle_vertices, land_cover_source, canopy_height_sou
         
         # Set default static height for trees (20 meters is a typical average tree height)
         static_tree_height = kwargs.get("static_tree_height", 10.0)
-        tree_mask = (land_cover_grid == 4)
+        # Determine tree class indices based on source-specific class names
+        _classes = get_land_cover_classes(land_cover_source)
+        _class_to_int = {name: i for i, name in enumerate(_classes.values())}
+        _tree_labels = ["Tree", "Trees", "Tree Canopy"]
+        _tree_indices = [_class_to_int[label] for label in _tree_labels if label in _class_to_int]
+        tree_mask = np.isin(land_cover_grid, _tree_indices) if _tree_indices else np.zeros_like(land_cover_grid, dtype=bool)
         
         # Set static height for tree cells
         canopy_height_grid_comp[tree_mask] = static_tree_height
+
+        # Bottom comp from trunk ratio
+        trunk_height_ratio = kwargs.get("trunk_height_ratio")
+        if trunk_height_ratio is None:
+            trunk_height_ratio = 11.76 / 19.98
+        canopy_bottom_height_grid_comp = canopy_height_grid_comp * float(trunk_height_ratio)
     else:
-        canopy_height_grid_comp = get_canopy_height_grid(rectangle_vertices, meshsize, canopy_height_source, output_dir, **kwargs)
+        canopy_height_grid_comp, canopy_bottom_height_grid_comp = get_canopy_height_grid(rectangle_vertices, meshsize, canopy_height_source, output_dir, **kwargs)
     
     # In the get_voxcity_CityGML function, modify it to handle None vegetation_gdf
     if vegetation_gdf is not None:
         canopy_height_grid = create_vegetation_height_grid_from_gdf_polygon(vegetation_gdf, meshsize, rectangle_vertices)
+        # Base bottom grid from ratio
+        trunk_height_ratio = kwargs.get("trunk_height_ratio")
+        if trunk_height_ratio is None:
+            trunk_height_ratio = 11.76 / 19.98
+        canopy_bottom_height_grid = canopy_height_grid * float(trunk_height_ratio)
     else:
         # Create an empty canopy_height_grid with the same shape as your other grids
         # This depends on the expected shape, you might need to adjust
         canopy_height_grid = np.zeros_like(building_height_grid)
+        canopy_bottom_height_grid = np.zeros_like(building_height_grid)
 
     mask = (canopy_height_grid == 0) & (canopy_height_grid_comp != 0)
     canopy_height_grid[mask] = canopy_height_grid_comp[mask]
+    # Apply same complementation to bottom grid
+    mask_b = (canopy_bottom_height_grid == 0) & (canopy_bottom_height_grid_comp != 0)
+    canopy_bottom_height_grid[mask_b] = canopy_bottom_height_grid_comp[mask_b]
+
+    # Ensure bottom <= top
+    canopy_bottom_height_grid = np.minimum(canopy_bottom_height_grid, canopy_height_grid)
     
     # Handle DEM - either flat or from source
     if kwargs.pop('flat_dem', None):
@@ -911,16 +1101,19 @@ def get_voxcity_CityGML(rectangle_vertices, land_cover_source, canopy_height_sou
     min_canopy_height = kwargs.get("min_canopy_height")
     if min_canopy_height is not None:
         canopy_height_grid[canopy_height_grid < kwargs["min_canopy_height"]] = 0        
+        canopy_bottom_height_grid[canopy_height_grid == 0] = 0
 
     # Remove objects near perimeter if specified
     remove_perimeter_object = kwargs.get("remove_perimeter_object")
     if (remove_perimeter_object is not None) and (remove_perimeter_object > 0):
+        print("apply perimeter removal")
         # Calculate perimeter width based on grid dimensions
         w_peri = int(remove_perimeter_object * building_height_grid.shape[0] + 0.5)
         h_peri = int(remove_perimeter_object * building_height_grid.shape[1] + 0.5)
         
         # Clear canopy heights in perimeter
         canopy_height_grid[:w_peri, :] = canopy_height_grid[-w_peri:, :] = canopy_height_grid[:, :h_peri] = canopy_height_grid[:, -h_peri:] = 0
+        canopy_bottom_height_grid[:w_peri, :] = canopy_bottom_height_grid[-w_peri:, :] = canopy_bottom_height_grid[:, :h_peri] = canopy_bottom_height_grid[:, -h_peri:] = 0
 
         # Find building IDs in perimeter regions
         ids1 = np.unique(building_id_grid[:w_peri, :][building_id_grid[:w_peri, :] > 0])
@@ -933,10 +1126,35 @@ def get_voxcity_CityGML(rectangle_vertices, land_cover_source, canopy_height_sou
         for remove_id in remove_ids:
             positions = np.where(building_id_grid == remove_id)
             building_height_grid[positions] = 0
-            building_min_height_grid[positions] = [[] for _ in range(len(building_min_height_grid[positions]))]    
+            building_min_height_grid[positions] = [[] for _ in range(len(building_min_height_grid[positions]))] 
+
+        # Visualize grids after optional perimeter removal
+        grid_vis = kwargs.get("gridvis", True)
+        if grid_vis:
+            # Building height grid visualization (zeros hidden)
+            building_height_grid_nan = building_height_grid.copy()
+            building_height_grid_nan[building_height_grid_nan == 0] = np.nan
+            visualize_numerical_grid(
+                np.flipud(building_height_grid_nan),
+                meshsize,
+                "building height (m)",
+                cmap='viridis',
+                label='Value'
+            )
+
+            # Canopy height grid visualization (zeros hidden)
+            canopy_height_grid_nan = canopy_height_grid.copy()
+            canopy_height_grid_nan[canopy_height_grid_nan == 0] = np.nan
+            visualize_numerical_grid(
+                np.flipud(canopy_height_grid_nan),
+                meshsize,
+                "Tree canopy height (m)",
+                cmap='Greens',
+                label='Tree canopy height (m)'
+            )
 
     # Generate 3D voxel grid
-    voxcity_grid = create_3d_voxel(building_height_grid, building_min_height_grid, building_id_grid, land_cover_grid, dem_grid, canopy_height_grid, meshsize, land_cover_source)
+    voxcity_grid = create_3d_voxel(building_height_grid, building_min_height_grid, building_id_grid, land_cover_grid, dem_grid, canopy_height_grid, meshsize, land_cover_source, canopy_bottom_height_grid)
 
     # Save all data if a save path is provided
     save_voxcity = kwargs.get("save_voxctiy_data", True)
@@ -946,7 +1164,7 @@ def get_voxcity_CityGML(rectangle_vertices, land_cover_source, canopy_height_sou
                          building_id_grid, canopy_height_grid, land_cover_grid, dem_grid, 
                          building_gdf, meshsize, rectangle_vertices)
     
-    return voxcity_grid, building_height_grid, building_min_height_grid, building_id_grid, canopy_height_grid, land_cover_grid, dem_grid, filtered_buildings
+    return voxcity_grid, building_height_grid, building_min_height_grid, building_id_grid, canopy_height_grid, canopy_bottom_height_grid, land_cover_grid, dem_grid, filtered_buildings
 
 def replace_nan_in_nested(arr, replace_value=10.0):
     """
